﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Net;
using System.Xml;

//XML return ids	message
//0                 (success)	
//1     			Internal failure
//2	    			Missing element
//3	    			Missing attribute
//4	    			Invalid interface name
//5	    			Invalid interface version
//6	    			Invalid request id
//7	    			Invalid module name
//8	    			Invalid method name
//9	    			Invalid argument name
//10	    		Invalid argument value
//11    			Invalid attribute value
//12    			Function failed

namespace HttpCtrlDemo
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        /*!
         * Request id can be used to pair up requests an their replies.
         */
        int reqId = 0;

        /*!
         * List of file paths selected for upload
         */
        List<string> uploadFileList = new List<string>();


        /*********************************************************************
         * FILE MENU ACTIONS:
         *********************************************************************/

        private void startUploadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            uploadFileInBackground();
        }

        private void stopUploadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            backgroundWorker1.CancelAsync();
        }

        /*********************************************************************
         * CONTROL MENU ACTIONS:
         *********************************************************************/

        private void playlistUploadDoneToolStripMenuItem_Click(object sender, EventArgs e)
        {
            var args = new Dictionary<string, string>();
            sendXmlRequestInBackground("control", "playlistUploadDone", args);
        }

        private void triggerId1ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sendNetworkTrigger("1");
        }

        private void triggerId2ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sendNetworkTrigger("2");
        }

        private void triggerId3ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sendNetworkTrigger("3");
        }

        private void triggerId4ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sendNetworkTrigger("4");
        }

        private void triggerId5ToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sendNetworkTrigger("5");
        }

        private void otherToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sendNetworkTrigger("");
        }

        private void sendNetworkTrigger(string id)
        {
            if (string.IsNullOrEmpty(id))
            {
                id = Microsoft.VisualBasic.Interaction.InputBox("Send network trigger", "Trigger Id:", "", 
                    this.Location.X + 100, this.Location.Y + 100);//TODO id = input user dialog + cancel
                int num;
                if(!int.TryParse(id, out num))
                {
                    return; //canceled or bad input
                }
            }
            var args = new Dictionary<string, string>();
            args.Add("triggerId", id);
            sendXmlRequestInBackground("control", "activatePlaylistTrigger", args);
        }


        /*********************************************************************
         * STATUS MENU ACTIONS:
         *********************************************************************/

        private void getPlayerStatusToolStripMenuItem_Click(object sender, EventArgs e)
        {
            var args = new Dictionary<string, string>();
            sendXmlRequestInBackground("status", "getPlayerStatus", args);
        }
        private void getMonitoringReportToolStripMenuItem_Click(object sender, EventArgs e)
        {
            var args = new Dictionary<string, string>();
            sendXmlRequestInBackground("status", "getMonitoringReport", args);
        }


        /*********************************************************************
         * HELP MENU ACTIONS:
         *********************************************************************/

        private void aboutHttpControlDemoToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AboutBox1 box = new AboutBox1();
            box.Show();
        }


        /*********************************************************************
         * FORM BUTTONS
         *********************************************************************/

        private void btnAddFiles_Click(object sender, EventArgs e)
        {
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                uploadFileList.AddRange(openFileDialog1.FileNames);
                lbxUploadQueue.DataSource = null;
                lbxUploadQueue.DataSource = uploadFileList;
                lblCurrentUpload.Text = "";
            }
        }

        private void btnClear_Click(object sender, EventArgs e)
        {
            uploadFileList.Clear();
            lbxUploadQueue.DataSource = null;
            lbxUploadQueue.DataSource = uploadFileList;
            lblCurrentUpload.Text = "";
        }


        /*********************************************************************
         * COMMON HTTP CTRL REQUEST FUNCTIONS
         *********************************************************************/
        
        private HttpWebRequest createHttpCtrlCgiRequest(string url, string userName, string password)
        {
            HttpWebRequest request = (HttpWebRequest)WebRequest.Create(url);
            request.Method = "POST";
            request.ServicePoint.Expect100Continue = false; //only needed for vipo3 version < 1.6.0
            //set request authorization
            string auth = String.Format("{0}:{1}", userName, password);
            byte[] authBytes = Encoding.ASCII.GetBytes(auth);
            string base64 = Convert.ToBase64String(authBytes);
            request.Headers.Add("Authorization", "Basic " + base64);
            return request;
        }

        /*!
         * Formats request content for triggering a playlist action on the device.
         * The trigger id value must be defined in the playlist that currently plays on the device,
         * in order to execute action associated with the trigger.
         * A MemoryStream and XmlWritter settings are used to switch from utf-16 encoding to utf-8 encoding.
         * Utf-16 encoded requests seems to work too, but we stick to the definition in httpctrl.
         * 
         * \param requestId - Holds the request id, a value which will be echoed back in the reply.
         * \param moduleName - defines the module
         * \param methodName - defines request type
         * \param args - optional arguments for 
         * 
         * \return an xml string to be sent as a content of a http request
        */
        private string formatRequestContent(string requestId, string moduleName, string methodName,
            Dictionary<string, string> args)
        {
            MemoryStream memoryStream = new MemoryStream();
            XmlWriterSettings xmlWriterSettings = new XmlWriterSettings();
            xmlWriterSettings.Encoding = new UTF8Encoding(false);
            xmlWriterSettings.ConformanceLevel = ConformanceLevel.Document;
            xmlWriterSettings.Indent = true;

            using (XmlWriter writer = XmlWriter.Create(memoryStream, xmlWriterSettings))
            {
                writer.WriteStartDocument();
                writer.WriteStartElement("interface");
                writer.WriteAttributeString("name", "httpctrl");
                writer.WriteAttributeString("version", "1.0.0");

                writer.WriteStartElement("request");
                writer.WriteAttributeString("id", requestId);
                writer.WriteEndElement(); //request

                writer.WriteStartElement("module");
                writer.WriteAttributeString("name", moduleName);

                writer.WriteStartElement("method");
                writer.WriteAttributeString("name", methodName);

                foreach (KeyValuePair<string, string> pair in args)
                {
                    writer.WriteStartElement("arg");
                    writer.WriteAttributeString("name", pair.Key);
                    writer.WriteAttributeString("value", pair.Value);
                    writer.WriteEndElement();
                }

            }//xml elements will be closed automatically in XmlWriter dispose

            return Encoding.UTF8.GetString(memoryStream.ToArray());
        }


        /*********************************************************************
         * FILE UPLOAD BACKGROUND WORKER:
         *********************************************************************/

        /*!
         * This method runs the file upload thread, so we dont block the gui. 
         * Large file upload can be a lenghty operation.
         */
        private void uploadFileInBackground()
        {
            if (uploadFileList.Count > 0)
            {
                string destFcgi = "httpctrl";
                
                BackGroundWorkerData data = new BackGroundWorkerData
                {
                    Url = String.Format("http://{0}/{1}", tbHost.Text, destFcgi), 
                    UserName = tbUserName.Text,
                    Password = tbPassword.Text,
                    LocalFilePath = uploadFileList.First()
                };

                btnAddFiles.Enabled = false;
                btnClear.Enabled = false;
                lblCurrentUpload.Text = "Start uploading...";
                backgroundWorker1.RunWorkerAsync(data);
            }
            else
            {
                lblCurrentUpload.Text = "No files to upload";
            }

        }

        /*!
         * This is done in the file uplaod thread (background worker1 thread)
         * */
        private void backgroundWorker1_DoWork(object sender, DoWorkEventArgs e)
        {
            //BackgroundWorker worker = sender as BackgroundWorker; can also get backgroundWorker1
            BackGroundWorkerData data = e.Argument as BackGroundWorkerData;

            FileInfo fi = new FileInfo(data.LocalFilePath);
            HttpWebRequest request = createHttpCtrlCgiRequest(data.Url, data.UserName, data.Password);
            request.Headers.Add("Content-Disposition",
                String.Format("method=\"setMediaFile\" filename=\"{0}\"", fi.Name));
            //set request content
            request.ContentType = "application/octet-stream";
            request.ContentLength = fi.Length;
            request.Timeout = 3600000;//1 hour
            request.KeepAlive = true;
            request.AllowWriteStreamBuffering = false;

            //open input and output streams
            Stream reqStream = request.GetRequestStream();
            FileStream fs = fi.OpenRead();

            //send content
            byte[] b = new byte[1400];
            int bytesRead = 0;
            long bytesDone = 0;
            int pctDone = 0;
            int oldPctDone = 0;

            while ((bytesRead = fs.Read(b, 0, b.Length)) != 0)
            {
                if ((backgroundWorker1.CancellationPending == true))
                {
                    e.Cancel = true;
                    request.Abort();
                    return;
                }
                else
                {
                    reqStream.Write(b, 0, bytesRead);
                    bytesDone += bytesRead;
                    pctDone = (int) (100 * bytesDone / fi.Length);

                    if (pctDone != oldPctDone)
                    {
                        backgroundWorker1.ReportProgress(pctDone);
                    }
                    oldPctDone = pctDone;
                }
            }

            fs.Close();

            reqStream.Close();

            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            e.Result = String.Format("Upload completed - {0}", fi.Name) + Environment.NewLine;
        }


        /*!
         * This handler updates the gui when the upload thread reports progress
         */
        private void backgroundWorker1_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            lblCurrentUpload.Text = (e.ProgressPercentage.ToString() + "% done");
            if (e.ProgressPercentage == 100)
            {
                lblCurrentUpload.Text += ", awaiting response...";
            }
        }

        /*!
         * BGW1 Run worker completed is called when the file upload thread exits.
         */
        private void backgroundWorker1_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            System.GC.Collect();
            btnAddFiles.Enabled = true;
            btnClear.Enabled = true;
            if ((e.Cancelled == true))
            {
                tbResponse.AppendText("Upload canceled!");
            }
            else if (e.Error != null)
            {
                tbResponse.AppendText("Error: " + e.Error.Message + Environment.NewLine);
                lblCurrentUpload.Text = "";
            }
            else
            {
                tbResponse.AppendText(e.Result as string);
                if (uploadFileList.Count > 0)
                {
                    uploadFileList.RemoveAt(0);
                    lbxUploadQueue.DataSource = null;
                    lbxUploadQueue.DataSource = uploadFileList;
                    uploadFileInBackground();
                }
            }
        }

        /*********************************************************************
        * XML REQUEST BACKGROUND WORKER:
        *********************************************************************/
        
        /*!
         * Send xml request and get response in the bacgkroundWorker2 thread.
         */
        private void sendXmlRequestInBackground(string moduleName, string methodName, Dictionary<string, string> args)
        {

            BackGroundWorkerData data = new BackGroundWorkerData
            {
                Url = String.Format("http://{0}/httpctrl", tbHost.Text),
                UserName = tbUserName.Text,
                Password = tbPassword.Text,
                RequestId = String.Format("DemoReq{0}", ++reqId),
                ModuleName = moduleName,
                MethodName = methodName,
                Arguments = args
            };

            if (backgroundWorker2.IsBusy)
            {
                tbResponse.AppendText(String.Format("'{0}' request ignored, BUSY!", methodName));
            }
            else
            {
                backgroundWorker2.RunWorkerAsync(data);
            }
        }
        
        /*!
         * When the thread runs, it send the request and waits for the reply.
         */
        private void backgroundWorker2_DoWork(object sender, DoWorkEventArgs e)
        {
            BackGroundWorkerData data = e.Argument as BackGroundWorkerData;
            HttpWebRequest request = createHttpCtrlCgiRequest(data.Url, data.UserName, data.Password);

            string xmlRequestContent = formatRequestContent(data.RequestId, data.ModuleName, data.MethodName, data.Arguments);

            //set request content
            request.ContentType = "text/xml";
            byte[] contentBytes = Encoding.ASCII.GetBytes(xmlRequestContent);
            request.ContentLength = contentBytes.Length;

            //send request
            Stream reqStream = request.GetRequestStream();
            reqStream.Write(contentBytes, 0, contentBytes.Length);
            reqStream.Close();

            HttpWebResponse response = (HttpWebResponse)request.GetResponse();
            Stream respStream = response.GetResponseStream();
            StreamReader respStreamReader = new StreamReader(respStream);

            //Get response as string. 
            //Read lines instead of whole strings to replace unix newline char with windows on the fly.
            //Response has only a couple of lines so string building is not an issue.

            string result = "";
            while (!respStreamReader.EndOfStream)
            {
                result += respStreamReader.ReadLine() + Environment.NewLine;
            }
            e.Result = result;
        }

        /*!
         * BGW2 Run worker completed is called when the xml request thread exits.
         */
        private void backgroundWorker2_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            if ((e.Cancelled == true))
            {
                tbResponse.AppendText("Canceled!");
            }
            else if (e.Error != null)
            {
                tbResponse.AppendText("Error: " + e.Error.Message + Environment.NewLine);
            }
            else
            {
                tbResponse.AppendText(e.Result as string);
            }
        }

        private void onClearResponseClick(object sender, EventArgs e)
        {
            tbResponse.Clear();
        }
    }
}

/*!
 * This class serves as a communication container for passing arguments to the background worker threads.
 */
class BackGroundWorkerData
{
    public string Url { get; set; }
    public string UserName { get; set; }
    public string Password { get; set; }
    public string LocalFilePath { get; set; }
    public string RequestId { get; set; }
    public string ModuleName { get; set; }
    public string MethodName { get; set; }
    public Dictionary<string, string> Arguments { get; set; }
}